# Copyright (c) 2018-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function

import copy
import os
import sys
from distutils import log
from distutils.dir_util import mkpath, remove_tree
from distutils.errors import DistutilsOptionError

import pkg_resources
from setuptools import Command
from xar import finders, pip_installer, py_util, xar_builder, xar_util


class bdist_xar(Command):
    description = "Create a XAR build distribution"

    user_options = [
        (
            "interpreter=",
            None,
            "The Python interpreter to use to run the XAR, which must be the ",
            "same version as sys.executable, default: a Python in the "
            "environment that is compatible with the running interpreter.",
        ),
        (
            "console-scripts=",
            None,
            "A common separated list of 'console_scripts' to build, or 'all'. "
            "Default: build the script with the package name, or if there is "
            "only one console script build that, otherwise fail.",
        ),
        ("download", None, "Download missing dependencies using pip"),
        (
            "xar-exec=",
            None,
            "Path to xarexec, which must be present to run the XAR file, "
            "default: /usr/bin/env xarexec_fuse.",
        ),
        (
            "xar-mount-root=",
            None,
            "Where the XAR file will be mounted by default, default: unset, "
            "will try both /mnt/xarfuse and /dev/shm.",
        ),
        (
            "xar-compression-algorithm=",
            None,
            "Compression algorithm for XAR file, default: gzip.",
        ),
        (
            "xar-block-size=",
            None,
            "Block size used when compressing the XAR file, default: 256K.",
        ),
        (
            "xar-zstd-level=",
            None,
            "Compression level when zstd compression is used, default: 16.",
        ),
        ("bdist-dir=", "b", "directory for building creating the distribution."),
        (
            "dist-dir=",
            "d",
            "directory for building creating the distribution, default: dist.",
        ),
        ("exclude-source-files", None, "remove all .py files from the generated egg"),
        (
            "keep-temp",
            "k",
            "keep the pseudo-installation tree around after "
            "creating the distribution archive",
        ),
        ("skip-build", None, "skip rebuilding everything (for testing/debugging)"),
    ]

    ALL = ["all"]

    boolean_options = ["exclude-source-files", "keep-temp", "skip-build"]

    def initialize_options(self):
        self.bdist_dir = None
        self.dist_dir = None
        self.exclude_source_files = False
        self.keep_temp = False
        self.skip_build = False
        self.console_scripts = None
        self.interpreter = None
        self.download = False
        # XAR options
        self.xar_exec = None
        self.xar_mount_root = None
        self.xar_compression_algorithm = None
        self.xar_block_size = None
        self.xar_zstd_level = None

    def finalize_options(self):
        if self.bdist_dir is None:
            bdist_base = self.get_finalized_command("bdist").bdist_base
            self.bdist_dir = os.path.join(bdist_base, "xar")
        if self.dist_dir is None:
            script_name = os.path.expanduser(self.distribution.script_name)
            package_dir = os.path.dirname(os.path.realpath(script_name))
            self.dist_dir = os.path.join(package_dir, "dist")
        if self.console_scripts is not None:
            self.console_scripts = self.console_scripts.strip().split(",")
        self.sqopts = xar_util.SquashfsOptions()
        if self.xar_compression_algorithm is not None:
            self.sqopts.compression_algorithm = self.xar_compression_algorithm
        else:
            self.sqopts.compression_algorithm = "gzip"
        if self.xar_block_size is not None:
            self.sqopts.block_size = self.xar_block_size
        if self.xar_zstd_level is not None:
            self.sqopts.zstd_level = self.xar_zstd_level
        self.xar_outputs = []

        self.working_set = pkg_resources.WorkingSet(sys.path)
        self.installer = None
        if self.download:
            bdist_pip = os.path.join(self.bdist_dir, "downloads")
            mkpath(bdist_pip)
            self.installer = pip_installer.PipInstaller(
                bdist_pip, self.working_set, log
            )

    def get_outputs(self):
        return self.xar_outputs

    def _add_distribution(self, xar):
        bdist_wheel = self.reinitialize_command("bdist_wheel")
        bdist_wheel.skip_build = self.skip_build
        bdist_wheel.keep_temp = self.keep_temp
        bdist_wheel.bdist_dir = os.path.join(self.bdist_dir, "wheel-bdist")
        bdist_wheel.dist_dir = os.path.join(self.bdist_dir, "wheel-dist")
        bdist_wheel.universal = False
        bdist_wheel.exclude_source_files = self.exclude_source_files
        bdist_wheel.distribution.dist_files = []
        self.run_command("bdist_wheel")
        assert len(bdist_wheel.distribution.dist_files) == 1
        wheel = bdist_wheel.distribution.dist_files[0][2]
        dist = py_util.Wheel(location=wheel).distribution
        xar.add_distribution(dist)
        return dist

    def _parse_console_scripts(self):
        """
        Get a map of console scripts to build based on :self.console_scripts:.
        """
        name = self.distribution.get_name()
        all_console_scripts = []
        entry_points = self.distribution.entry_points
        if entry_points:
            entry_points = pkg_resources.EntryPoint.parse_map(entry_points)
            all_console_scripts = entry_points.get("console_scripts", {})
        if len(all_console_scripts) == 0:
            raise DistutilsOptionError("'%s' has no 'console_scripts'" % name)
        if self.console_scripts == self.ALL:
            return all_console_scripts
        if self.console_scripts is None:
            if len(all_console_scripts) == 1:
                return all_console_scripts
            if name in all_console_scripts:
                return {name: all_console_scripts[name]}
            raise DistutilsOptionError(
                "More than one entry point, set --console-scripts"
            )
        console_scripts = {}
        for script in self.console_scripts:
            if script not in all_console_scripts:
                raise DistutilsOptionError("'%s' is not in 'console_scripts'" % script)
            console_scripts[script] = all_console_scripts[script]
        return console_scripts

    def _set_entry_point(self, xar, entry_point):
        attrs = ".".join(entry_point.attrs)
        entry_point_str = "%s:%s" % (entry_point.module_name, attrs)
        xar.set_entry_point(entry_point_str)

    def _deps(self, dist, extras=()):
        requires = dist.requires(extras=extras)
        try:
            finders.register_finders()
            # Requires setuptools>=34.1 for the bug fix.
            return set(
                self.working_set.resolve(
                    requires, extras=extras, installer=self.installer
                )
            )
        except pkg_resources.DistributionNotFound:
            name = self.distribution.get_name()
            requires_str = "\n\t".join(str(req) for req in requires)
            log.error(
                "%s's requirements are not satisfied:\n\t%s\n"
                "Either pass --download to bdist_xar to download missing "
                "dependencies with pip or try 'pip install /path/to/%s'."
                % (name, requires_str, name)
            )
            raise

    def _build_entry_point(self, base_xar, dist, common_deps, entry_name, entry_point):
        # Clone the base xar
        xar = copy.deepcopy(base_xar)
        # Add in any extra dependencies
        deps = self._deps(dist, entry_point.extras)
        deps -= common_deps
        for dep in deps:
            log.info("adding dependency '%s' to xar" % dep.project_name)
            xar.add_distribution(dep)
        # Set the entry point
        self._set_entry_point(xar, entry_point)
        # Build the XAR
        xar_output = os.path.join(self.dist_dir, entry_name + ".xar")
        mkpath(self.dist_dir)
        log.info("creating xar '%s'" % xar_output)
        xar.build(xar_output, self.sqopts)
        self.xar_outputs.append(xar_output)

    def run(self):
        try:
            xar = xar_builder.PythonXarBuilder(self.xar_exec, self.xar_mount_root)
            # Build an egg for this package and import it.
            dist = self._add_distribution(xar)
            # Add in the dependencies common to each entry_point
            deps = self._deps(dist)
            for dep in deps:
                log.info("adding dependency '%s' to xar" % dep.project_name)
                xar.add_distribution(dep)
            # Set the interpreter to the current python interpreter
            if self.interpreter is not None:
                xar.set_interpreter(self.interpreter)
            # Build a XAR for each entry point specified
            entry_points = self._parse_console_scripts()
            for entry_name, entry_point in entry_points.items():
                self._build_entry_point(xar, dist, deps, entry_name, entry_point)
        finally:
            # Clean up the build directory
            if not self.keep_temp:
                remove_tree(self.bdist_dir)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
"""
A simple script to create a mountable archive file (XAR file), from an existing
Python executable archive, or a directory.

Usage (basic XAR):
  make_xar --raw=dir_to_xarify \
    --output=/path/to/output.xar

Usage (executable XAR):
  make_xar --xar_exec=/path/to/xarexec \
    --raw-executable=executable_inside_squash_file \
    --output=/path/to/output_executable

Usage (Python XAR):
  make_xar --python=dir_or_archive_to_xarify \
    --python-entry-point=module.in.squash_file:main_function

The script is designed to turn any directory or existing Python executable
archive into a xar file. It creates the squashfs file and the shebang that
invokes xarexec (if requested). The xar_builder API is available for more
involved use cases.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import logging
import os
import sys
import zipfile

from xar import py_util, xar_builder, xar_util


class XarArgumentError(Exception):
    pass


def main(args=None):
    logging.basicConfig(
        level=logging.DEBUG, format="%(asctime)s %(levelname)s %(message)s"
    )
    p = argparse.ArgumentParser()
    p.add_argument("--output", required=True, help="Output XAR file.")
    # XAR options
    p.add_argument(
        "--xar-exec",
        help="Path to xarexec, which must be present to run the XAR file.",
        default="/usr/bin/env xarexec_fuse",
    )
    p.add_argument(
        "--xar-mount-root",
        default=None,
        help="Where the XAR file will mount by default.",
    )

    p.add_argument(
        "--xar-compression-algorithm",
        default="gzip",
        help="Compression algorithm for the XAR file.",
    )
    p.add_argument(
        "--xar-block-size",
        default=256 * 1024,
        help="Block size used when compressing the XAR file.",
    )
    p.add_argument(
        "--xar-zstd-level",
        default=16,
        help="Default zstd level when zstd compression is used.",
    )

    group = p.add_mutually_exclusive_group(required=True)
    # Python options
    group.add_argument(
        "--python",
        help="Make an executable python XAR from the given" "directory or zip archive.",
    )
    p.add_argument(
        "--python-interpreter",
        default=None,
        help="Python interpreter for building and running the XAR. "
        "If not set and constructing from a zip archive we try "
        "to extract the shebang to get the Python interpreter. "
        "Otherwise we default to an interpreter compatible with the running Python.",
    )
    p.add_argument(
        "--python-entry-point",
        default=None,
        help="MODULE[:FUNCTION]"
        "The entry point for the python XAR. If unset, we look "
        "for a __main__ module in the XAR and use that.",
    )
    # Raw options
    group.add_argument("--raw", help="Make a raw xar from a directory")
    p.add_argument(
        "--raw-executable",
        default=None,
        help="Executable invoked once the XAR is mounted. It must "
        "be a path relative to the XAR root. If unset the XAR "
        "is not executable. The arguments passed to the executable when "
        "invoked are the XAR path, followed by all the arguments passed to "
        "the XAR.",
    )
    opts = p.parse_args(args)

    squashfs_options = xar_util.SquashfsOptions()
    squashfs_options.compression_algorithm = opts.xar_compression_algorithm
    squashfs_options.block_size = opts.xar_block_size
    squashfs_options.zstd_level = opts.xar_zstd_level

    if opts.python:
        xar = xar_builder.PythonXarBuilder(opts.xar_exec, opts.xar_mount_root)
        interpreter = opts.python_interpreter
        entry_point = opts.python_entry_point
        # Either copy the directory or extract the archive.
        # Infer interpreter and entry_point if unset.
        if os.path.isdir(opts.python):
            xar.add_directory(opts.python)
            entry_point = entry_point or py_util.get_python_main(opts.python)
        else:
            z_interpreter, z_entry_point = py_util.extract_python_archive_info(
                opts.python
            )
            with zipfile.ZipFile(opts.python) as zf:
                interpreter = interpreter or z_interpreter
                entry_point = entry_point or z_entry_point
                xar.add_zipfile(zf)
        if entry_point is None:
            raise XarArgumentError("Python entry point not set and no __main__")

        if interpreter is not None:
            xar.set_interpreter(interpreter)
        xar.set_entry_point(entry_point)
    elif opts.raw:
        xar = xar_builder.XarBuilder(opts.xar_exec, opts.xar_mount_root)
        xar.add_directory(opts.raw)
        if opts.raw_executable is not None:
            xar.set_executable(opts.raw_executable)
    else:
        raise ValueError("Unexpected value")

    xar.build(opts.output, squashfs_options)

    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import base64
import csv
import hashlib
import logging
import os
import platform
import py_compile
import re
import shlex
import shutil
import subprocess
import sys
import tempfile
import zipfile
import zipimport

import pkg_resources
from xar import xar_util
from xar.compat import cache_from_source, native, source_from_cache
from xar.vendor.wheel import install, paths, pkginfo


logger = logging.getLogger("xar")


PYTHON_EXTS = [".py", ".pyc", ".pyo"]


def parse_entry_point(entry_point):
    """
    Parses a Python entry point and returns the module and function.
    The two allowed formats are 'path.to.module', and 'path.to.module:function'.
    In the former case, ('path.to.module', None) is returned.
    In the latter case, ('path.to.module', 'function') is returned.
    """
    module, sep, function = entry_point.partition(":")
    if function and sep and module:
        return (module, function)
    else:
        return (module, None)


def get_python_main(directory):
    """Returns the python __main__ from a directory (if it exists)."""
    main = os.path.join(directory, "__main__")
    main_exists = any(os.path.exists(main + ext) for ext in PYTHON_EXTS)
    if main_exists:
        return "__main__"
    return None


def extract_python_archive_info(archive):
    """
    Extracts the shebang (if any) from a python archive, along with the entry
    point (if any). Returns a tuple (python_interpreter, entry_point).
    Avoids interpreting the shebang in it doesn't contain 'python'.
    """
    python = None
    with open(archive, "rb") as f:
        if f.read(2) == b"#!":
            shebang = f.readline().decode("utf-8").strip()
            if "python" in shebang:
                python = shebang
    with zipfile.ZipFile(archive) as zf:
        # Ignores __pycache__ since .pyc in __pycache__ aren't executable
        # without the .py.
        MAIN = "__main__"
        main_exists = any(xar_util.file_in_zip(zf, MAIN + ext) for ext in PYTHON_EXTS)
        if main_exists:
            return (python, MAIN)
        return (python, None)


def get_pyc_file(py_file):
    """Returns the .pyc file for `py_file` (not .pyo)."""
    return cache_from_source(py_file, debug_override=True)


def compile_files(py_files):
    """
    Compiles every Python file in `py_files` into a pyc file.
    The pyc file name is determined by `get_pyc_file(py_file)`.
    Returns a dict of files that errored to the error message.
    Note: Always writes to .pyc, even if optimization is enabled.
    """
    errors = {}
    for py_file in py_files:
        pyc_file = get_pyc_file(py_file)
        try:
            py_compile.compile(py_file, pyc_file, doraise=True)
            assert os.path.exists(pyc_file)
            newtime = xar_util.extract_pyc_timestamp(pyc_file)
            os.utime(py_file, (newtime, newtime))
        except py_compile.PyCompileError as e:
            errors[py_file] = e.msg
    return errors


def is_python_version(interpreter, version_info):
    """
    Returns `True` if `interpreter` is version `version_info`.
    """
    VERSION_INFO_MAIN = """
import sys
print('.'.join(str(x) for x in sys.version_info))
    """
    assert interpreter is not None
    with tempfile.NamedTemporaryFile("w+t", delete=False) as f:
        f.write(VERSION_INFO_MAIN)
        f.flush()
        this_version = ".".join(str(x) for x in version_info)
        binary = shlex.split(interpreter)
        output = subprocess.check_output(binary + [f.name])
        other_version = output.decode("utf-8").strip()
        return this_version == other_version


# Cribbed from PythonIdentity.hashbang() in PEX.
def environment_python_interpreter():
    """
    Returns an Python interpreter found in the enviornment that is compatible
    with the currently running Python interpreter.
    """
    ENV_PYTHON = {
        "CPython": "python%(major)d.%(minor)d",
        "Jython": "jython",
        "PyPy": "pypy",
        "IronPython": "ipy",
    }
    python = ENV_PYTHON[platform.python_implementation()] % {
        "major": sys.version_info[0],
        "minor": sys.version_info[1],
        "patch": sys.version_info[2],
    }
    return "/usr/bin/env %s" % python


class WheelMetadata(pkg_resources.EggMetadata):
    """Metadata provider for zipped wheels."""

    def _setup_prefix(self):
        # Cribbed from pkg_resources.EggProvider and pex WheelMetadata.
        for path in xar_util.yield_prefixes_reverse(self.module_path):
            if Wheel.is_wheel_archive(path):
                self.egg_name = os.path.basename(path)
                wf = install.WheelFile(path)
                self.egg_info = os.path.join(path, wf.distinfo_name)
                self.egg_root = path
                break


def does_sha256_match(file, expected_hash):
    """
    Does `file`'s sha256 match `expected_hash`. The `expected_hash` is expected
    to be in the format of RECORD files: sha256=urlsafe_b64_with_no_trailing_==.
    """
    h = hashlib.sha256()
    with open(file, "rb") as f:
        data = f.read(4096)
        while data:
            h.update(data)
            data = f.read(4096)
    hash = b"sha256=" + base64.urlsafe_b64encode(h.digest()).rstrip(b"=")
    return native(hash) == expected_hash


class Wheel(object):
    """
    Wrapper around a pkg_resources.DistInfoDistribution with Wheel specific
    helpers.
    """

    class Error(Exception):
        pass

    WHEEL_INFO = install.WheelFile.WHEEL_INFO
    RECORD = install.WheelFile.RECORD

    # The latest wheel no longer accepts "package-version.dist-info".
    # Copy the old regex here.
    # See https://github.com/pypa/wheel/issues/236
    WHEEL_INFO_RE = re.compile(
        r"""^(?P<namever>(?P<name>.+?)(-(?P<ver>\d.*?))?)
        ((-(?P<build>\d.*?))?-(?P<pyver>.+?)-(?P<abi>.+?)-(?P<plat>.+?)
        \.whl|\.dist-info)$""",
        re.VERBOSE,
    ).match

    @classmethod
    def is_wheel_archive(cls, path):
        """Returns True if `path` is a wheel."""
        return path.lower().endswith(".whl")

    def __init__(self, distribution=None, location=None, importer=None):
        """
        Constructs the WheelDistribution
        """
        if distribution is not None and location is not None:
            raise self.Error("location and distribution cannot both be set")

        if distribution is not None:
            self.distribution = distribution
        else:
            # Construct the metadata provider
            if self.is_wheel_archive(location):
                importer = importer or zipimport.zipimporter(location)
                metadata = WheelMetadata(importer)
            else:
                root = os.path.dirname(location)
                metadata = pkg_resources.PathMetadata(root, location)
            project_name, version, py_version, platform = [None] * 4
            match = self.WHEEL_INFO_RE(os.path.basename(metadata.egg_info))
            if match:
                project_name, version, py_version, platform = match.group(
                    "name", "ver", "pyver", "plat"
                )
                py_version = py_version or sys.version_info[0]
            self.distribution = pkg_resources.DistInfoDistribution(
                location,
                metadata,
                project_name=project_name,
                version=version,
                py_version=py_version,
                platform=platform,
            )
        # self.distribution.egg_info is the only reliable way to get the name.
        # I'm not sure if egg_info is a public interface, but we already rely
        # on it for WheelMetadata.
        wheel_info = os.path.basename(self.distribution.egg_info)
        parsed_filename = self.WHEEL_INFO_RE(wheel_info)
        if parsed_filename is None:
            raise self.Error("Bad wheel '%s'" % wheel_info)
        self.name, self.ver, self.namever = parsed_filename.group(
            "name", "ver", "namever"
        )

    def is_purelib(self):
        """Returns True if the Wheel is a purelib."""
        wheel_info = pkginfo.read_pkg_info_bytes(
            self.distribution.get_metadata(self.WHEEL_INFO).encode("utf-8")
        )
        return wheel_info["Root-Is-Purelib"] == "true"

    def records(self):
        """
        Returns an iterator over the records of the Wheel.
        Iterates over triples [filename, hash, size].
        """
        return csv.reader(self.distribution.get_metadata_lines(self.RECORD))

    def distinfo_name(self):
        return "%s.dist-info" % self.namever

    def distinfo_location(self, install_paths):
        """
        Returns the location of the distinfo using the `install_paths`.
        The `Wheel` can be reconstructed from this location.
        """
        if self.is_purelib():
            root = install_paths["purelib"]
        else:
            root = install_paths["platlib"]
        return os.path.join(root, self.distinfo_name())

    def sys_install_paths(self):
        """Return the system wheel install locations."""
        sys_paths = paths.get_install_paths(self.name)
        for key in sys_paths:
            sys_paths[key] = os.path.realpath(sys_paths[key])
        return sys_paths

    def install(self, src_paths, dst_paths, force=False):
        """
        Install this Wheel distribution to `dst_paths`.
        Uses either :func:`install_archive` or :func:`copy_installation`.
        """
        if self.is_wheel_archive(self.distribution.location):
            self.install_archive(dst_paths, force)
        else:
            self.copy_installation(src_paths, dst_paths, force)

    def install_archive(self, dst_paths, force=False):
        """
        Install a Wheel archive to `dst_paths`.
        """
        if not self.is_wheel_archive(self.distribution.location):
            raise self.Error("install() only works with archives")
        wf = install.WheelFile(self.distribution.location)
        # TODO: The next version of wheel doesn't provide WheelFile,
        # so we will need to have our own implementation of wf.install().
        # When you fix that, please make force=False allow overwrites if the
        # hashes match. You can use does_sha256_match().
        wf.install(overrides=dst_paths, force=True)

    def _determine_kind(self, src_root, src_paths, dst_paths, src_record):
        """
        Determine the most specific `src_paths` kind that the `src_record` is
        located under. If the most specific kind has the same `src_paths[kind]`
        as another kind, then the `dst_paths` must be the same as well.
        """
        kinds = []
        for prefix in xar_util.yield_prefixes_reverse(src_record):
            if prefix in src_paths.values():
                kinds = [kind for kind, path in src_paths.items() if prefix == path]
                break
        else:
            # We were unable to determine the kind, fall back to a heuristic.
            # * If the path contains "/lib/" it is a {pure,plat}lib.
            # * If the path contains "/bin/" it is a script.
            # * If the path contains the name and "include" it is a header.
            # * If the src_root ends with "site-pacakges", the data may be in
            #   "src_root/../../..".
            # This is necessary for Python installed with Homebrew on OS X.
            data = None
            if src_root.endswith("site-packages"):
                data = os.path.normpath(os.path.join(src_root, "../../.."))
            for prefix in xar_util.yield_prefixes_reverse(src_record):
                if prefix.endswith("site-packages"):
                    assert dst_paths["purelib"] == dst_paths["platlib"]
                    return "purelib", prefix
                if prefix.endswith("bin"):
                    return "scripts", prefix
                if prefix.endswith(self.name) and "include" in prefix:
                    return "headers", prefix
                if data and prefix == data:
                    return "data", prefix
        # We must have exactly one unique prefix
        xar_prefix = [dst_paths[kind] for kind in kinds]
        if not kinds or not all(p == xar_prefix[0] for p in xar_prefix):
            raise self.Error(
                "Distribution '%s' has record '%s' with ambiguous kind."
                % (self.namever, src_record)
            )
        kind = kinds[0]
        return kind, src_paths[kind]

    def copy_installation(self, src_paths, dst_paths, force=False):
        """
        Copies an installation of this Wheel in `src_paths` to `dst_paths`.
        Takes care to fix up the 'RECORD' file to reflect the new installtion
        location.
        """
        if self.is_wheel_archive(self.distribution.location):
            raise self.Error("copy_installation() does not work with archives")
        # Determine the src and dst roots.
        if self.is_purelib():
            src_root = src_paths["purelib"]
            dst_root = dst_paths["purelib"]
        else:
            src_root = src_paths["platlib"]
            dst_root = dst_paths["platlib"]
        assert src_root[:1] == os.sep and dst_root[:1] == os.sep
        # Create or overwrite the dst RECORD file.
        dst_records_path = os.path.join(dst_root, self.distinfo_name(), self.RECORD)
        if os.path.exists(dst_records_path) and not force:
            raise self.Error("'RECORD' already exists: '%s'" % self.name)
        xar_util.safe_mkdir(os.path.dirname(dst_records_path))
        with open(dst_records_path, mode="w+t") as f:
            # Loop over each record in the source distribution.
            dst_records = csv.writer(f)
            for record, record_hash, record_size in self.records():
                # Get the normalized absolute path for the source record
                src_record = os.path.normpath(os.path.join(src_root, record))
                # Determine what 'kind' the record is, and get dst_record path.
                kind, prefix = self._determine_kind(
                    src_root, src_paths, dst_paths, src_record
                )
                rel_record = src_record[len(prefix) + 1 :]
                dst_record = os.path.join(dst_paths[kind], rel_record)
                # Update the destination RECORD file.
                new_record = os.path.relpath(dst_record, dst_root)
                dst_records.writerow((new_record, record_hash, record_size))
                # Don't write the records file, since we are recreating it.
                if dst_record == dst_records_path:
                    continue
                # Copy or overwrite the record
                if os.path.exists(dst_record) and not force:
                    if not does_sha256_match(dst_record, record_hash):
                        raise self.Error("'%s' already exists" % dst_record)
                xar_util.safe_mkdir(os.path.dirname(dst_record))
                shutil.copy2(src_record, dst_record)

    def fixup(self, install_paths):
        """
        Deletes .pyc files if the .py file exists (to be recompiled).
        Compiles all Python sources.
        """
        root = os.path.dirname(self.distinfo_location(install_paths))
        # Read the RECORDS
        records = list(self.records())
        # Get a list of Python files (and a set for quick membership tests)
        py_files = [
            os.path.normpath(os.path.join(root, record))
            for record, _, _ in records
            if record.endswith(".py")
        ]
        py_set = set(py_files)
        # Write the new RECORDS file, excluding .pyc files
        new_records = []
        for record_line in records:
            record, _, _ = record_line
            # Add non-pyc files and continue
            is_py = record.endswith(".py")
            if is_py or not any(record.endswith(e) for e in PYTHON_EXTS):
                new_records.append(record_line)
                continue
            # We have a pyc file, delete it if the .py file exists.
            pyc_file = os.path.normpath(os.path.join(root, record))
            try:
                py_file = source_from_cache(pyc_file)
                if py_file in py_set:
                    xar_util.safe_remove(pyc_file)
                    continue
            except ValueError:
                pass
            # There is no .py file, add the .pyc file
            new_records.append(record_line)
        # Compile all the py files
        # Failures are okay since there might be for example Python 3 only code
        errors = compile_files(py_files)
        for _, msg in errors.items():
            logger.warning(msg)
        # Add the new .pyc files relative to root
        new_records += [
            [os.path.relpath(get_pyc_file(f), root), "", ""]
            for f in py_files
            if f not in errors
        ]
        # Write the new RECORDS file
        records_path = os.path.join(root, self.distinfo_name(), self.RECORD)
        with open(records_path, "wt") as f:
            writer = csv.writer(f)
            writer.writerows(new_records)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Handy helper utils to work with XARs at runtime"""

from __future__ import absolute_import, division, print_function, unicode_literals

import os


def get_runtime_path():
    # type: (...) -> str
    """Return the location of the runtime files directory"""

    runtime_path = os.getenv("XAR_RUNTIME_FILES")
    if runtime_path:
        if not os.access(runtime_path, os.R_OK):
            raise ValueError("XAR_RUNTIME_FILES is invalid: %s" % runtime_path)
        return runtime_path

    raise ValueError("Cannot determine runtime files path.")



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
"""
A trivial XAR used to confirm XARs execute correctly on a host.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
from pathlib import Path


# We don't really use pyinit but this ensures we can import native
# code properly, which indicates the XAR is properly mounted, etc.
# For now, this is only for Linux; macs don't support native code.
if sys.platform == "linux":
    from libfb.py import pyinit

    parser = pyinit.FbcodeArgumentParser()
    opts = parser.parse_args(sys.argv[1:])

# Print any env variables that are XAR or PAR related.
for k, v in sorted(os.environ.items()):
    if k.startswith(("FB_XAR", "FB_PAR")):
        print("%s=%s" % (k, v))

assert "FB_XAR_INVOKED_NAME" in os.environ
binary_name = os.path.splitext(os.path.basename(os.getenv("FB_XAR_INVOKED_NAME")))[0]

for env in (
    "FB_PAR_RUNTIME_FILES",
    # 'LD_LIBRARY_PATH',
    # 'LD_PRELOAD',
):
    assert env in os.environ, "%s not in environment" % env

xar_mountpoint = Path(os.environ["FB_PAR_RUNTIME_FILES"])
for file in ("xar_bootstrap.sh", "libtools_xar_%s-cxx-build-info-lib.so" % binary_name):
    assert os.access(xar_mountpoint / file, os.R_OK), "%s isn't accessible" % file

if sys.platform == "linux":
    with open("/proc/self/maps") as maps_file:
        maps = maps_file.read()
    for file in ("libtools_xar_%s-cxx-build-info-lib.so" % binary_name,):
        assert str(xar_mountpoint / file) in maps, "%s not preloaded" % file

print("ok")



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import os

from xar import xar_builder, xar_util
from xar.tests import xar_test_helpers


class XarBuilderTest(xar_test_helpers.XarTestCase):
    def setUp(self):
        self.xar_builder = xar_builder.XarBuilder()
        self.sqopts = xar_util.SquashfsOptions()
        self.sqopts.compression_algorithm = "gzip"
        self.src = xar_util.StagingDirectory()
        # A set of files for testing
        self.files = {
            "executable.sh": ("executable.sh", "#!echo executable", "w", 0o755),
            "lib.so": ("lib.so", b"binary source", "wb", 0o644),
            "source.txt": ("source.txt", "text source", "w", 0o644),
            "subdir/source.txt": ("subdir/source.txt", "subdir", "w", 0o644),
        }
        for filename, data, mode, permissions in self.files.values():
            self.src.write(data, filename, mode, permissions)
            self.assertEqual(
                xar_test_helpers.mode(self.src.absolute(filename)), permissions
            )

    def tearDown(self):
        self.xar_builder.delete()
        self.src.delete()

    def _set_xar_builder(self, xar_builder):
        self.xar_builder.delete()
        self.xar_builder = xar_builder

    def _src_file(self, key):
        return self.src.absolute(self.files[key][0])

    def _staging(self):
        return self.xar_builder._staging

    def test_add_file(self):
        # Add a file
        src_file = self._src_file("source.txt")
        self.xar_builder.add_file(src_file)
        dst_file = self.xar_builder._staging.absolute("source.txt")
        self.assertFilesEqual(src_file, dst_file)
        # Try to add the file again
        with self.assertRaises(xar_util.StagingDirectory.Error):
            self.xar_builder.add_file(src_file)
        # Add an executable in a subdirectory
        src_file = self._src_file("executable.sh")
        self.xar_builder.add_file(src_file, "subdir/file")
        dst_file = self.xar_builder._staging.absolute("subdir/file")
        self.assertFilesEqual(src_file, dst_file)
        # Try to add after frozen
        self.xar_builder.freeze()
        with self.assertRaises(xar_builder.XarBuilder.FrozenError):
            self.xar_builder.add_file(src_file, "file")

    def test_add_directory(self):
        # Set the staging dir to the test directory
        self.xar_builder.add_directory(self.src.path())
        self.assertDirectoryEqual(self.src.path(), self._staging().path())
        # Add a copy to a subdir
        self.xar_builder.add_directory(self.src.path(), "my-subdir")
        subdir = os.path.join(self._staging().path(), "my-subdir")
        self.assertDirectoryEqual(self.src.path(), subdir)
        # Attempt to overwrite the staging directory
        with self.assertRaises(xar_util.StagingDirectory.Error):
            self.xar_builder.add_directory(self.src.path())
        # Attempt to overwrite the subdir
        with self.assertRaises(xar_util.StagingDirectory.Error):
            self.xar_builder.add_directory(self.src.path(), "my-subdir")
        # Try to add after frozen
        self.xar_builder.freeze()
        with self.assertRaises(xar_builder.XarBuilder.FrozenError):
            self.xar_builder.add_directory(self.src.path(), "another-subdir")

    def test_set_shebang(self):
        # Invalid shebangs
        with self.assertRaises(xar_builder.XarBuilder.InvalidShebangError):
            self.xar_builder._set_shebang("bang")
        with self.assertRaises(xar_builder.XarBuilder.InvalidShebangError):
            self.xar_builder._set_shebang("#!too long" + ("0" * 130))
        # Valid shebang
        self.xar_builder._set_shebang("#!bang")
        self.assertEqual("#!bang", self.xar_builder._shebang)
        # Shebang already set
        with self.assertRaises(xar_builder.XarBuilder.InvalidShebangError):
            self.xar_builder._set_shebang("#!bang")
        # Try to set after frozen
        self.xar_builder.freeze()
        with self.assertRaises(xar_builder.XarBuilder.FrozenError):
            self.xar_builder._set_shebang("#!bang")

    def test_set_executable(self):
        # Invalid executable
        with self.assertRaises(xar_builder.XarBuilder.InvalidExecutableError):
            self.xar_builder.set_executable("bad")
        # Valid executable
        src = self._src_file("executable.sh")
        self.xar_builder.add_executable(src)
        self.assertEqual("#!/usr/bin/env xarexec_fuse", self.xar_builder._shebang)
        self.assertFilesEqual(src, self._staging().absolute("executable.sh"))
        # Executable already set
        with self.assertRaises(xar_builder.XarBuilder.InvalidExecutableError):
            self.xar_builder.add_executable(src, "different")
        # Try to add after frozen
        self.xar_builder.freeze()
        with self.assertRaises(xar_builder.XarBuilder.FrozenError):
            self.xar_builder.set_executable("executable")

    def test_sort(self):
        self.xar_builder.add_directory(self.src.path())
        self.xar_builder.sort_by_extension([".txt", ".so", ""])
        # Try to override
        with self.assertRaises(xar_builder.XarBuilder.Error):
            self.xar_builder.sort_by_extension([".txt", ".so", ""])
        # Override
        self.xar_builder.sort_by_extension([".txt", ".so", ""], override=True)

        self.xar_builder.freeze()
        # Try to set while frozen
        with self.assertRaises(xar_builder.XarBuilder.FrozenError):
            self.xar_builder.sort_by_extension(None, override=True)
        # Check the result
        self.assertTrue(self.xar_builder._sort_file is not None)
        with self.xar_builder._sort_file.open("r") as f:
            sort_data = f.read().strip()
        sort_data = [l.split(" ") for l in sort_data.split("\n")]
        for filename, priority in sort_data:
            if filename.endswith(".txt"):
                self.assertEqual(priority, "-4")
            elif filename.endswith(".so"):
                self.assertEqual(priority, "-3")
            else:
                self.assertEqual(priority, "-2")

    def test_partition(self):
        self.xar_builder.add_directory(self.src.path())
        self.xar_builder.partition_by_extension([".txt"])
        # Try to override
        with self.assertRaises(xar_builder.XarBuilder.Error):
            self.xar_builder.partition_by_extension([".txt"])
        # Override
        self.xar_builder.partition_by_extension([".txt"], override=True)

        self.xar_builder.freeze()
        # Try to set while frozen
        with self.assertRaises(xar_builder.XarBuilder.FrozenError):
            self.xar_builder.partition_by_extension(None, override=True)
        # Check the source directory
        for dirpath, _dirnames, filenames in os.walk(self._staging().path()):
            for filename in filenames:
                abs_filename = os.path.join(dirpath, filename)
                if filename.endswith(".txt"):
                    self.assertTrue(os.path.islink(abs_filename))
                else:
                    self.assertFalse(os.path.islink(abs_filename))
        # Check the partion directory
        self.assertEqual(len(self.xar_builder._partition_dest), 1)
        dest = self.xar_builder._partition_dest[".txt"]
        for dirpath, _dirname, filenames in os.walk(dest[0].path()):
            for filename in filenames:
                abs_filename = os.path.join(dirpath, filename)
                self.assertTrue(filename.endswith(".txt"))
                self.assertFalse(os.path.islink(abs_filename))
        # The directories should have the same file names
        self.assertDirectoryEqual(
            self.src.path(), self._staging().path(), check_contents=False
        )

    def test_build(self):
        self.xar_builder.add_directory(self.src.path())
        dst = xar_util.StagingDirectory()
        test_xar = os.path.join(dst.path(), "test.xar")
        test_root = os.path.join(dst.path(), "squashfs-root")
        self.xar_builder.build(test_xar, self.sqopts)
        self._unxar(test_xar, test_root)
        self.assertDirectoryEqual(self.src.path(), test_root)
        dst.delete()

    def test_sort_build(self):
        self.xar_builder.add_directory(self.src.path())
        self.xar_builder.sort_by_extension([".txt", ".so", ""])
        dst = xar_util.StagingDirectory()
        test_xar = os.path.join(dst.path(), "test.xar")
        test_root = os.path.join(dst.path(), "squashfs-root")
        self.xar_builder.build(test_xar, self.sqopts)
        self._unxar(test_xar, test_root)
        self.assertDirectoryEqual(self.src.path(), test_root)
        dst.delete()

    def test_deepcopy(self):
        other = copy.deepcopy(self.xar_builder)
        self.assertEqual(other._sort_file, None)
        self.assertEqual(other._frozen, self.xar_builder._frozen)
        self.assertEqual(other._xar_exec, self.xar_builder._xar_exec)
        other.delete()

        self.xar_builder._sort_file = xar_util.TemporaryFile()
        other = copy.deepcopy(self.xar_builder)
        self.assertNotEqual(
            other._staging.absolute(), self.xar_builder._staging.absolute()
        )
        self.assertNotEqual(other._sort_file.name(), self.xar_builder._sort_file.name())
        other.delete()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil
import sys
import tempfile
import time

from xar import bootstrap_py, py_util, xar_util


BORING_SHEBANG = "#!/bin/echo This is not an executable XAR file."
MAX_SHEBANG = 128  # from linux/include/linux/binfmts.h's BINPRM_BUF_SIZE


class XarBuilder(object):
    """
    Handles the construction of a XAR.
    Files and directories can be added to the XAR until it is frozen, then it
    cannot be modified, and can only be built with XarBuilder.build().
    """

    class Error(Exception):
        pass

    class InvalidExecutableError(Error):
        pass

    class FrozenError(Error):
        pass

    class InvalidShebangError(Error):
        pass

    def __init__(self, xar_exec=None, mount_root=None):
        """
        Constructs a XarBuilder given the optional `xar_exec` path and
        `mount_root`. `xar_exec` is the executable that runs the xar and
        `mount_root` is where `xar_exec` will mount the squashfs filesystem.
        """
        self._staging = xar_util.StagingDirectory()
        self._frozen = False

        self._mount_root = mount_root
        self._xar_exec = xar_exec
        if self._xar_exec is None:
            self._xar_exec = "/usr/bin/env xarexec_fuse"

        self._executable = None
        self._shebang = None
        self._priorities = None
        self._partition = None

        self._version = None
        self._sort_file = None
        self._partition_dest = {}

    def _ensure_frozen(self):
        if not self._frozen:
            raise self.FrozenError("Expected to be frozen")

    def _ensure_unfrozen(self):
        if self._frozen:
            raise self.FrozenError("Expected not to be frozen")

    def add_file(self, filename, xar_filename=None):
        """
        Adds `filename` to the XAR with the xar filename `xar_filename`, which
        does not have to match. If `xar_filename` is `None`, then the basename
        of `filename` is used. Throws if the file already exists.
        """
        self._ensure_unfrozen()
        if xar_filename is None:
            xar_filename = os.path.basename(filename)
        self._staging.copy(filename, xar_filename)

    def add_directory(self, directory, xar_directory=None):
        """
        Adds `directory` to the XAR with the xar directory name `xar_directory`,
        which does not have to match. If `xar_directory` is `None`, then the
        root of the xar is set to the directory. Throws if the directory already
        exists.
        """
        self._ensure_unfrozen()
        self._staging.copytree(directory, xar_directory)

    def add_zipfile(self, zf, dst=None):
        """
        Adds a zipfile to the XAR under `dst`. If `dst` is `None`, then the root
        of the xar is set to the directory. Throws if any extracted file already
        exists.
        """
        self._ensure_unfrozen()
        self._staging.extract(zf, dst)

    def _set_shebang(self, shebang):
        """Sets the shebang."""
        self._ensure_unfrozen()
        if self._shebang is not None:
            raise self.InvalidShebangError("Already have a shebang")
        if not shebang.startswith("#!"):
            raise self.InvalidShebangError("Invalid shebang '%s'" % shebang)
        if len(shebang) > MAX_SHEBANG:
            raise self.InvalidShebangError("Shebang too long '%s'" % shebang)
        self._shebang = shebang

    def set_executable(self, xar_filename):
        """
        Sets the executable to the relative path `xar_filename`, which must
        already be present in the XAR. `xar_exec` will execute this file when
        the XAR is executed.
        """
        self._ensure_unfrozen()
        if self._executable is not None:
            raise self.InvalidExecutableError("Already have an executable")
        if not self._staging.exists(xar_filename):
            raise self.InvalidExecutableError(
                "Executable '%s' does not exist" % xar_filename
            )
        self._set_shebang("#!%s" % self._xar_exec)
        self._executable = xar_filename

    def add_executable(self, filename, xar_filename=None):
        """Calls :func:`add_file` and then :func:`set_executable`."""
        if xar_filename is None:
            xar_filename = os.path.basename(filename)
        self.add_file(filename, xar_filename)
        self.set_executable(xar_filename)

    def sort_by_extension(self, priorities, override=False):
        """
        Sets the sort priorities for mksquashfs. `priorities` is a list of file
        extensions. Files are sorted in the order that their extensions appear
        in `priorities`. This may be called before adding all the files, since
        the work will only happen once the XarBuilder is frozen. Refuses to
        override existing priorities unless `override` is `True`.
        """
        self._ensure_unfrozen()
        if not override and self._priorities is not None:
            raise self.Error("Refusing to override existing sort priorities")
        self._priorities = priorities

    def _run_sort_by_extension(self):
        """Performs the :func:`sort_by_extension` work once frozen."""
        self._ensure_frozen()
        self._sort_file = None
        if self._priorities is None:
            return
        self._sort_file = xar_util.TemporaryFile()
        with self._sort_file.open(mode="w+t") as f:
            xar_util.write_sort_file(self._staging.path(), self._priorities, f)

    def partition_by_extension(self, partition, override=False):
        """
        Partitions the files into separate XARs based on their extension.
        `partition` is a list of extensions which go into their own XARs.
        The partitioned XARs are added as dependencies to the main XAR.
        This may be called before adding all the files, since the work will only
        happen once the XarBuilder is frozen. Refuses to override existing
        partition unless `override` is True.
        """
        self._ensure_unfrozen()
        if not override and self._partition is not None:
            raise self.Error("Refusing to override existing ext. partition")
        self._partition = partition

    def _run_partition_by_extension(self):
        """Performs the :func:`partition_by_extension` work once frozen."""
        self._ensure_frozen()
        self._partition_dest = {}
        if self._partition is None:
            return
        for ext in self._partition:
            staging_dir = xar_util.StagingDirectory()
            uuid = xar_util.make_uuid()
            dest = xar_util.PartitionDestination(staging_dir, uuid)
            self._partition_dest["." + ext.lstrip(".")] = dest
        xar_util.partition_files(self._staging, self._partition_dest)

    def freeze(self):
        """
        Freezes the XarBuilder. After this point the XarBuilder may not be
        modified, and the only valid operations are :func:`build` and
        :func:`delete`.
        """
        self._ensure_unfrozen()
        if self._shebang is None:
            self._set_shebang(BORING_SHEBANG)
        self._frozen = True
        self._version = int(time.time())
        self._run_sort_by_extension()
        self._run_partition_by_extension()

    def delete(self):
        """
        Delete temporary resources. Only necessary if :func:`build` is never
        called. The XarBuilder is no longer usable after this call.
        """
        self._staging.delete()
        if self._sort_file is not None:
            self._sort_file.delete()
        for dest in self._partition_dest.values():
            dest.staging.delete()

    def _build_staging_dir(
        self, staging_dir, filename, shebang, xar_header, squashfs_options
    ):
        """Builds a single XAR."""
        self._ensure_frozen()
        os.chmod(staging_dir.path(), 0o755)
        xar = xar_util.XarFactory(staging_dir.path(), filename, shebang)
        if xar_header is not None:
            xar.xar_header = xar_header.copy()
        xar.version = self._version
        xar.version = self._version
        if self._sort_file:
            xar.sort_file = self._sort_file.name()
        xar.squashfs_options = squashfs_options
        xar.go()

    def _build_xar_header(self, xar_dependencies):
        """Make the XAR headers."""
        self._ensure_frozen()
        xar_header = {}
        xar_header["DEPENDENCIES"] = " ".join(
            [os.path.basename(v[0]) for v in xar_dependencies]
        )
        if self._mount_root:
            xar_header["MOUNT_ROOT"] = self._mount_root
        if self._executable is not None:
            xar_header["XAREXEC_TARGET"] = self._executable
        return xar_header

    def build(self, filename, squashfs_options=None):
        """
        Actually build the XAR. Freezes the XarBuilder if not already frozen.
        Writes the XAR to `filename`. Uses `xar_util.SquashfsOptions`
        `squashfs_options` to construct the XAR. Finally calls :func:`delete`.
        The XarBuilder is no longer usable after this call.
        """
        if squashfs_options is None:
            squashfs_options = xar_util.SquashfsOptions()
        if not self._frozen:
            self.freeze()
        xarfiles = {}
        base_name, xar_ext = os.path.splitext(filename)
        # Build the dependent XARs
        for ext, destination in self._partition_dest:
            ext_filename = base_name + ext + xar_ext
            with tempfile.NamedTemporaryFile(delete=False) as tf:
                xarfiles[ext] = (ext_filename, tf.name)
            self._build_staging_dir(
                destination.staging,
                xarfiles[ext][1],
                BORING_SHEBANG,
                {},
                squashfs_options,
            )
        # Build the main XAR
        with tempfile.NamedTemporaryFile(delete=False) as tf:
            tmp_xar = tf.name
        xar_header = self._build_xar_header(xarfiles)
        self._build_staging_dir(
            self._staging, tmp_xar, self._shebang, xar_header, squashfs_options
        )

        # Move the results into place
        shutil.move(tmp_xar, filename)
        for ext_filename, tmp_filename in xarfiles.values():
            shutil.move(tmp_filename, ext_filename)

        # Make the output executable if necessary
        if self._executable is not None:
            os.chmod(filename, 0o755)

        self.delete()


class PythonXarBuilder(XarBuilder):
    """
    Handles the construction of a Python executable XAR.
    All of the Python dependencies should be bundled into the XAR, since the XAR
    will set the `$PYTHONPATH` to be the mounted squash filesystem during the
    bootstrap phase. The invoked executable name is stored in the environment
    variable `$XAR_INVOKED_NAME` and the mounted XAR squashfs filesystem root is
    stored in the environment variable `$XAR_RUNTIME_FILES`. Extra dependencies
    can be added with the XarBuilder methods. The XAR will execute the given
    Python entry point with the given Python interpreter.
    """

    class InvalidEntryPointError(XarBuilder.Error):
        pass

    class InvalidInterpreterError(XarBuilder.Error):
        pass

    class InvalidDistributionError(XarBuilder.Error):
        pass

    LIBRARY_PATH = ""

    def __init__(self, *args, **kwargs):
        self._entry_point = None
        self._interpreter = None
        self._distributions = set()

        super(PythonXarBuilder, self).__init__(*args, **kwargs)

    def _validate_entry_point(self, entry_point):
        """Validates that the module specified in `entry_point` exists."""

        def ensure_exists(module):
            basename = os.path.join(self.LIBRARY_PATH, *module.split("."))
            if os.path.isdir(self._staging.absolute(basename)):
                basename = basename + "/__init__"
            for ext in py_util.PYTHON_EXTS:
                if self._staging.exists(basename + ext):
                    return
            raise self.InvalidEntryPointError("Module '%s' not found in XAR" % module)

        module, function = py_util.parse_entry_point(entry_point)

        parent_end = len(module)
        while parent_end > 0:
            parent_module = module[:parent_end]
            ensure_exists(parent_module)
            parent_end = parent_module.rfind(".")

    def set_entry_point(self, entry_point):
        """
        Sets the Python entry point for the XAR. In the format `module:function`
        or `module`. This is what the XAR executes after bootstrapping. Must be
        called before :func:`freeze`.
        """
        self._ensure_unfrozen()
        self._validate_entry_point(entry_point)
        if self._entry_point is not None:
            raise self.InvalidEntryPointError("Entry point is already set")
        self._entry_point = entry_point

    def set_interpreter(self, interpreter):
        """
        Sets the python interpreter to `python`. Defaults to calling
        :func:`py_util.environment_python_interpreter` for the interpreter if
        not set.
        """
        self._ensure_unfrozen()
        if self._interpreter is not None:
            raise self.InvalidInterpreterError("Interpreter is already set")
        if not py_util.is_python_version(interpreter, sys.version_info):
            raise self.InvalidInterpreterError(
                "%s is not compatible with the running Python version." % interpreter
            )
        self._interpreter = interpreter

    def _xar_install_paths(self, dist_name, absolute):
        """Return the XAR wheel install locations."""
        prefix = ""
        if absolute:
            prefix = self._staging.absolute() + os.sep
        return {
            "purelib": "%s%s" % (prefix, self.LIBRARY_PATH),
            "platlib": "%s%s" % (prefix, self.LIBRARY_PATH),
            "headers": "%sinclude/%s" % (prefix, dist_name),
            "scripts": "%sbin" % prefix,
            "data": prefix,
        }

    def add_distribution(self, distribution):
        """
        Add a `pkg_resources.Distribution` to the XAR. The distribution must be
        a wheel, but it may be either a zipfile or already unpacked.
        Handles all the installation, and adding the distribution to the Python
        path.
        """
        self._ensure_unfrozen()
        # We only support wheels.
        if not distribution.has_metadata(py_util.Wheel.WHEEL_INFO):
            raise self.InvalidDistributionError(
                "'%s' is not a wheel! It might be an egg, try reinstalling as "
                "a wheel." % distribution.project_name
            )
        wheel = py_util.Wheel(distribution=distribution)
        sys_paths = wheel.sys_install_paths()
        xar_paths = self._xar_install_paths(wheel.name, absolute=True)
        wheel.install(sys_paths, xar_paths, force=False)
        self._distributions.add(wheel.distinfo_location(xar_paths))

    def _fixup_distributions(self):
        """Fixup the distributions."""
        for distinfo_location in self._distributions:
            wheel = py_util.Wheel(location=distinfo_location)
            xar_paths = self._xar_install_paths(wheel.name, absolute=True)
            wheel.fixup(xar_paths)

    def _bootstrap(self):
        """Set up the Python bootstrapping."""
        if self._interpreter is None:
            self._interpreter = py_util.environment_python_interpreter()
        if self._entry_point is None:
            raise self.InvalidEntryPointError("Entry point is not set")

        module, function = py_util.parse_entry_point(self._entry_point)
        fmt_args = {
            "python": self._interpreter,
            "module": module,
            "run_xar_main": bootstrap_py.RUN_XAR_MAIN,
        }
        if function is not None:
            fmt_args["function"] = function
        bootstrap_xar = bootstrap_py.BOOTSTRAP_XAR_TEMPLATE.format(**fmt_args)
        run_xar_main = bootstrap_py.run_xar_main(**fmt_args)

        self._staging.write(
            bootstrap_xar, bootstrap_py.BOOTSTRAP_XAR, mode="w", permissions=0o755
        )
        self._staging.write(
            run_xar_main, bootstrap_py.RUN_XAR_MAIN, mode="w", permissions=0o644
        )
        self.set_executable(bootstrap_py.BOOTSTRAP_XAR)

    def freeze(self):
        """
        See :func:`XarBuilder.freeze`. Adds in the extra Python specific steps.
        """
        self._fixup_distributions()
        self._bootstrap()
        super(PythonXarBuilder, self).freeze()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import contextlib
import errno
import logging
import os
import platform
import shutil
import stat
import struct
import subprocess
import sys
import tempfile
import time
import uuid


logger = logging.getLogger("xar")

if "centos" in platform.platform().lower():
    NOGROUP = "nobody"
else:
    # Works for debian and darwin for sure
    NOGROUP = "nogroup"


def make_uuid():
    # ugh line length limit; we need a small uuid
    return str(uuid.uuid1()).split("-")[0]


def _align_offset(offset, align=4096):
    """Aligns the offset to the given alignment"""
    mask = align - 1
    assert (mask & align) == 0
    return (offset + mask) & (~mask)


class SquashfsOptions(object):
    def __init__(self):
        self.compression_algorithm = "zstd"
        self.zstd_level = 16
        self.block_size = 256 * 1024


class XarFactory(object):
    """A class for creating XAR files.

    Pretty straight forward; take an input directory, output file, and some
    metadata and produce a XAR file of the contents.
    """

    def __init__(self, dirname, output, header_prefix):
        self.dirname = dirname
        self.output = output
        self.header_prefix = header_prefix
        self.xar_header = {}
        self.uuid = None
        self.version = None
        self.sort_file = None
        self.squashfs_options = SquashfsOptions()

    def go(self):
        "Make the XAR file."
        logger.info("Squashing %s to %s" % (self.dirname, self.output))
        if self.uuid is None:
            self.uuid = make_uuid()

        if self.version is None:
            self.version = time.time()

        tf = tempfile.NamedTemporaryFile(delete=False)
        # Create!
        sqopts = self.squashfs_options
        cmd = [
            "mksquashfs",
            self.dirname,
            tf.name,
            "-noappend",
            "-noI",
            "-noX",  # is this worth it?  probably
            "-force-uid",
            "nobody",
            "-force-gid",
            NOGROUP,
            "-b",
            str(sqopts.block_size),
            "-comp",
            sqopts.compression_algorithm,
        ]
        if sqopts.compression_algorithm == "zstd":
            cmd.extend(("-Xcompression-level", str(sqopts.zstd_level)))

        if self.sort_file:
            cmd.extend(["-sort", self.sort_file])

        if sys.stdout.isatty():
            subprocess.check_call(cmd)
        else:
            with open("/dev/null", "wb") as f:
                subprocess.check_call(cmd, stdout=f)

        headers = [self.header_prefix]
        # Take the squash file, create a header, and write it
        with open(self.output, "wb") as of:
            # Make a "safe" header that is easily parsed and also not
            # going to explode if accidentally executed.
            headers.append('OFFSET="$OFFSET"')
            headers.append('UUID="$UUID"')
            headers.append('VERSION="%d"' % self.version)
            for key, val in self.xar_header.items():
                headers.append('%s="%s"' % (key, str(val).replace('"', " ")))
            headers.append("#xar_stop")
            headers.append("echo This XAR file should not be executed by sh")
            headers.append("exit 1")
            headers.append("# Actual squashfs file begins at $OFFSET")
            text_headers = "\n".join(headers) + "\n"
            # 128 is to account for expansion of $OFFSET and $UUID;
            # it's well over what they might reasonably be.
            header_size = _align_offset(128 + len(text_headers))
            text_headers = text_headers.replace("$OFFSET", "%d" % header_size)
            text_headers = text_headers.replace("$UUID", self.uuid)
            text_headers += "\n" * (header_size - len(text_headers))
            of.write(text_headers.encode("UTF-8"))

            # Now append the squashfs file to the header.
            with open(tf.name, "rb") as rf:
                while True:
                    data = rf.read(1024 * 1024)
                    if not data:
                        break
                    of.write(data)


def safe_mkdir(directory):
    try:
        os.makedirs(directory)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise


def safe_remove(filename):
    try:
        os.unlink(filename)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def safe_rmtree(directory):
    if os.path.exists(directory):
        shutil.rmtree(directory, True)


# Simplified version of Chroot from PEX
class StagingDirectory(object):
    """
    Manages the staging directory.
    """

    class Error(Exception):
        pass

    def __init__(self, staging_dir=None):
        self._staging = os.path.normpath(staging_dir or tempfile.mkdtemp())
        safe_mkdir(self._staging)

    def __deepcopy__(self, memo):
        other = StagingDirectory()
        memo[id(self)] = other
        other.copytree(self._staging)
        return other

    def _normalize(self, dst):
        dst = os.path.normpath(dst)
        if dst.startswith(os.sep) or dst.startswith(".."):
            raise self.Error("Destination path '%s' is not a relative!" % dst)
        return dst

    def _ensure_parent(self, dst):
        safe_mkdir(os.path.dirname(self.absolute(dst)))

    def _ensure_not_dst(self, dst):
        if self.exists(dst):
            raise self.Error("Destination path '%s' already exists!" % dst)

    def path(self):
        """Returns the root directory of the staging directory."""
        return self._staging

    def absolute(self, dst=None):
        """Returns absolute path for a path relative to staging directory."""
        if dst is None:
            return self._staging
        dst = self._normalize(dst)
        return os.path.normpath(os.path.join(self._staging, dst))

    def delete(self):
        """Delete the staging directory."""
        safe_rmtree(self._staging)

    def copy(self, src, dst):
        """Copy src into dst under the staging directory."""
        dst = self._normalize(dst)
        self._ensure_parent(dst)
        self._ensure_not_dst(dst)
        shutil.copy2(src, self.absolute(dst))

    def write(self, data, dst, mode, permissions):
        """Write data into dst."""
        dst = self._normalize(dst)
        self._ensure_parent(dst)
        self._ensure_not_dst(dst)
        with open(self.absolute(dst), mode) as f:
            f.write(data)
        os.chmod(self.absolute(dst), permissions)

    @contextlib.contextmanager
    def postprocess(self, src):
        fpath = self.absolute(src)
        st = os.stat(fpath)
        old_times = (st.st_atime, st.st_mtime)

        with tempfile.NamedTemporaryFile(
            prefix=fpath + ".", mode="w", delete=False
        ) as outf:
            with open(fpath) as inf:
                yield inf, outf

            outf.flush()
            os.utime(outf.name, old_times)
            shutil.copystat(fpath, outf.name)
            os.rename(outf.name, fpath)

    def _resolve_dst_dir(self, dst):
        if dst is None:
            # Replace the current staging directory
            if os.listdir(self._staging) != []:
                raise self.Error("Staging directory is not empty!")
            # shutil requires that the destination directory does not exist
            safe_rmtree(self._staging)
            dst = "."
        dst = self._normalize(dst)
        self._ensure_not_dst(dst)
        return dst

    def copytree(self, src, dst=None):
        """Copy src dir into dst under the staging directory."""
        dst = self._resolve_dst_dir(dst)
        shutil.copytree(src, self.absolute(dst))

    def symlink(self, link, dst):
        """Write symbolic link to dst under the staging directory."""
        dst = self._normalize(dst)
        self._ensure_parent(dst)
        self._ensure_not_dst(dst)
        os.symlink(link, self.absolute(dst))

    def move(self, src, dst):
        """Move src into dst under the staging directory."""
        dst = self._normalize(dst)
        self._ensure_parent(dst)
        self._ensure_not_dst(dst)
        shutil.move(src, self.absolute(dst))

    def exists(self, dst):
        """Checks if dst exists under the staging directory."""
        dst = self._normalize(dst)
        return os.path.exists(self.absolute(dst))

    def extract(self, zf, dst=None):
        """Extracts the zipfile into dst under the staging directory."""
        dst = self._resolve_dst_dir(dst)
        abs_dst = os.path.join(self._staging, dst)
        timestamps = {}
        for zi in zf.infolist():
            filename = os.path.join(dst, zi.filename)
            destination = self.absolute(filename)

            mode = zi.external_attr >> 16
            if stat.S_ISLNK(mode):
                target = zf.read(zi).decode("utf-8")
                self.symlink(target, filename)
            else:
                self._ensure_parent(filename)
                zf.extract(zi, path=abs_dst)
                os.chmod(destination, stat.S_IMODE(mode))

            # Use the embedded timestamp for from the pyc file for the
            # pyc and py file; otherwise, use the timezone-less
            # timestamp from the zipfile (sigh).
            if filename.endswith(".pyc"):
                new_time = extract_pyc_timestamp(destination)
                timestamps[destination] = new_time  # pyc file
                timestamps[destination[:-1]] = new_time  # py file too
            else:
                new_time = tuple((list(zi.date_time) + [0, 0, -1]))
                timestamps[destination] = time.mktime(new_time)

        # Set our timestamps.
        for path, timestamp in timestamps.items():
            try:
                os.utime(path, (timestamp, timestamp))
            except OSError as e:
                # Sometimes we had a pyc file but no py file; the utime
                # would fail.
                if not path.endswith(".py"):
                    raise e


class TemporaryFile(object):
    """Wrapper around a temporary file that supports deepcopy()."""

    def __init__(self):
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
            self._filename = f.name

    def open(self, mode=None):
        return open(self._filename, mode)

    def name(self):
        return self._filename

    def delete(self):
        safe_remove(self._filename)

    def __deepcopy__(self, memo):
        other = TemporaryFile()
        memo[id(self)] = other
        with self.open("rb") as src, other.open("wb") as dst:
            shutil.copyfileobj(src, dst)
        return other


# Simple class to represent a partition destination.  Each destination
# is a path and a uuid from which the contents come (ie, the uuid of
# the spar file that contains the file that is moved into the
# partition; used for symlink construction).
PartitionDestination = collections.namedtuple("PartitionDestination", "staging uuid")


def partition_files(staging, extension_destinations):
    """Partition source_dir into multiple output directories.

    A partition is defined by extension_destinations which maps suffixes (such
    as ".debuginfo") to a PartitionDestination instance.

    dest_dir contains all files that aren't in a partition, and symlinks for
    ones that are.  symlinks are relative and of the form
    "../../../uuid/path/to/file" so that the final symlinks are correct
    relative to /mnt/xar/....
    """
    source_dir = staging.path()
    source_dir = source_dir.rstrip("/")

    for dirpath, _dirnames, filenames in os.walk(staging.path()):
        # path relative to source_dir; used for creating the right
        # file inside the staging dir
        relative_dirname = dirpath[len(source_dir) + 1 :]

        # Special case; if a file is in the root of source_dir, then
        # relative_dirname is empty, but that has the same number of
        # '/' as just 'bin', so we need to special case it the empty
        # value.
        if not relative_dirname:
            relative_depth = 1
        else:
            relative_depth = 2 + relative_dirname.count("/")

        for filename in filenames:
            # Does this extension map to a separate output?
            _, extension = os.path.splitext(filename)
            dest_base = extension_destinations.get(extension, None)
            # This path stays in the source staging directory
            if dest_base is None:
                continue
            # This file is destined for another tree, make a
            # relative symlink in source pointing to the
            # sub-xar destination.
            relative_path = os.path.join(relative_dirname, filename)
            source_path = staging.absolute(relative_path)
            dest_base.staging.move(source_path, relative_path)

            dependency_mountpoint = dest_base.uuid
            staging_symlink = os.path.join(
                "../" * relative_depth, dependency_mountpoint, relative_path
            )
            logging.info("%s %s" % (staging_symlink, source_path))

            staging.symlink(staging_symlink, relative_path)


def write_sort_file(staging_dir, extension_priorities, sort_file):
    """
    Write a sort file for mksquashfs to colocate some files at the beginning.
    Files are assigned priority by extension, with files earlier in the list
    appearing first. The result is written to the file object sort_file.
    mksquashfs takes the sort file with the option '-sort sort_filename'.
    """
    for dirpath, _dirname, filenames in os.walk(staging_dir):
        for filename in filenames:
            fn = os.path.join(dirpath, filename)
            for idx, suffix in enumerate(extension_priorities):
                if fn.endswith(suffix):
                    # Default priority is 0; make ours all
                    # negative so we can not list files with
                    # spaces in the name, making them default
                    # to 0
                    priority = idx - len(extension_priorities) - 1
                    break

            assert fn.startswith(staging_dir + "/")
            fn = fn[len(staging_dir) + 1 :]

            # Older versions of mksquashfs don't like spaces
            # in filenames; let them have the default priority
            # of 0.
            if " " not in fn:
                sort_file.write("%s %d\n" % (fn, priority))


def extract_pyc_timestamp(path):
    "Extract the embedded timestamp from a pyc file"

    # A PYC file has a four byte header then four byte timestamp.  The
    # timestamp must match the timestamp on the py file, otherwise the
    # interpreter will attempt to re-compile the py file.  We extract
    # the timestamp to adulterate the py/pyc files before squashing
    # them.
    with open(path, "rb") as fh:
        prefix = fh.read(8)
        return struct.unpack(b"<I", prefix[4:])[0]


def file_in_zip(zf, filename):
    """Returns True if :filename: is present in the zipfile :zf:."""
    try:
        zf.getinfo(filename)
        return True
    except KeyError:
        return False


def yield_prefixes_reverse(path):
    """
    Yields all prefixes of :path: in reverse.
    list(yield_prefixes_reverse("/a/b")) == ["/a/b", "/a", "/"]
    list(yield_prefixes_reverse("a/b")) == ["a/b", "a", ""]
    """
    old = None
    while path != old:
        yield path
        old = path
        path, _ = os.path.split(path)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import subprocess
import tempfile
import unittest

from xar import xar_builder


def mode(filename):
    return os.stat(filename).st_mode & 0o777


class XarTestCase(unittest.TestCase):
    def _unxar(self, xarfile, outdir):
        """unsquashfs the xarfile into the outdir."""
        # Make sure the header is what we expect; also grab the offset.
        with open(xarfile, "rb") as fh:
            first_line = fh.readline()
            shebang = first_line.decode("utf-8").strip()
            self.assertEquals(shebang, xar_builder.BORING_SHEBANG)
            saw_stop = False
            offset = None
            for line in fh:
                if line == b"#xar_stop\n":
                    saw_stop = True
                    break
                if line.startswith(b"OFFSET="):
                    offset = int(line[8:-2])
            self.assertTrue(saw_stop)
            self.assertTrue(offset % 4096 == 0)

            fh.seek(offset)
            squashfs_contents = fh.read()

        # Write the squashfs file out, expand it, and make sure it
        # contains the same files as the source.
        with tempfile.NamedTemporaryFile() as out, open("/dev/null", "wb") as devnull:
            out.write(squashfs_contents)
            out.flush()
            subprocess.check_call(
                ["unsquashfs", "-d", outdir, "-no-xattrs", out.name], stdout=devnull
            )

    def assertDirectoryEqual(self, src, dst, check_contents=True):
        """Verify two directories contain the same entries, recursively."""

        def directory_contents(d):
            ret = []
            for dirname, dirs, files in os.walk(d):
                for entry in dirs + files:
                    full_path = os.path.join(dirname, entry)
                    ret.append(full_path[len(d) + 1 :])
            return sorted(ret)

        src_contents = directory_contents(src)
        dst_contents = directory_contents(dst)
        self.assertEqual(src_contents, dst_contents)
        for src_file, dst_file in zip(src_contents, dst_contents):
            src_file = os.path.join(src, src_file)
            dst_file = os.path.join(dst, dst_file)
            if check_contents and os.path.isfile(src_file):
                self.assertFilesEqual(src_file, dst_file)

    def assertFilesEqual(self, src, dst):
        """Verify that the contents of two files are the same."""
        self.assertTrue(os.path.exists(src))
        self.assertTrue(os.path.exists(dst))
        with open(src, "rb") as fh:
            src_contents = fh.read()
        with open(dst, "rb") as fh:
            dst_contents = fh.read()
        self.assertEqual(src_contents, dst_contents)
        self.assertEqual(mode(src), mode(dst))



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from __future__ import absolute_import, division, print_function, unicode_literals

import tempfile
import unittest
from shutil import rmtree

import xar.utils


try:
    import unittest.mock as mock
except ImportError:
    import mock


class XarUtilsTestCase(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        try:
            rmtree(self.tempdir)
        except OSError:
            pass

    def test_get_runtime_path(self):
        with mock.patch("xar.utils.os.getenv") as fake_env:
            fake_env.return_value = self.tempdir
            self.assertEqual(xar.utils.get_runtime_path(), self.tempdir)

            fake_env.return_value = None
            with self.assertRaises(ValueError):
                xar.utils.get_runtime_path()


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function

import contextlib
import multiprocessing
import os
import subprocess
import sys
import tarfile
import tempfile
import zipfile

import pkg_resources
from xar import finders, py_util, xar_util


class PipException(Exception):
    pass


class BuildException(Exception):
    pass


class PipInstaller(object):
    """
    Installer function object for pkg_resources.working_set.resolve().
    It is called like `installer(requirement)` when the requirement can't be
    found in the working set. See `__call__()` for documentation.
    """

    def __init__(self, dest, working_set, log=None):
        """
        Download Wheels to `dest` and add them to the `working_set`.
        """
        self._dest = dest
        self._working_set = working_set
        self._log = log
        self._working_set.add_entry(self._dest)
        req = pkg_resources.Requirement.parse("pip")
        dist = self._working_set.find(req)
        self._pip_main = pkg_resources.load_entry_point(dist, "console_scripts", "pip")

    def clean(self):
        """
        Remove any non-wheels from the downloads directory.
        """
        for entry in os.listdir(self._dest):
            if not py_util.Wheel.is_wheel_archive(entry):
                xar_util.safe_remove(os.path.join(self._dest, entry))

    def invoke_pip(self, args):
        def main(args):
            sys.exit(self._pip_main(args))

        p = multiprocessing.Process(target=main, args=(args,))
        p.start()
        p.join()
        if p.exitcode == 0:
            return
        raise PipException("'pip %s' failed" % " ".join(args))

    def download(self, requirement):
        """
        Download the requirement to the downloads directory using pip.
        """
        args = ["download", "-d", self._dest, str(requirement)]
        self.invoke_pip(args)

    def extract_sdist(self, sdist, dest):
        """
        Extract the sdist archive and return the path to the source.
        """
        if sdist.lower().endswith(".zip"):
            open_sdist = zipfile.ZipFile
            error_cls = zipfile.BadZipfile
        else:
            assert ".tar" in sdist.lower()
            open_sdist = tarfile.TarFile.open
            error_cls = tarfile.ReadError
        try:
            with contextlib.closing(open_sdist(sdist)) as archive:
                archive.extractall(path=dest)

            def collapse_trivial(path):
                entries = os.listdir(path)
                if len(entries) == 1:
                    entry = os.path.join(path, entries[0])
                    if os.path.isdir(entry):
                        return collapse_trivial(entry)
                return path

            return collapse_trivial(dest)
        except error_cls:
            raise BuildException("Failed to extract %s" % os.path.basename(sdist))

    def build_wheel_from_sdist(self, sdist):
        """
        Given a sdist archive extract it, build in a temporary directory, and
        put the wheel into the downloads directory.
        """
        temp = tempfile.mkdtemp()
        try:
            source = self.extract_sdist(sdist, temp)
            # Make sure to import setuptools and wheel in the setup.py.
            # This is happening in a temporary directory, so we will just
            # overwrite the setup.py to add our own imports.
            setup_py = os.path.join(source, "setup.py")
            with open(setup_py, "r") as f:
                original = f.read()
            with open(setup_py, "w") as f:
                f.write("import setuptools\n")
                f.write("import wheel\n")
                f.write(original)
            # Build the wheel
            command = [
                sys.executable,
                setup_py,
                "bdist_wheel",
                "-d",
                os.path.abspath(self._dest),
            ]
            subprocess.check_call(command, cwd=source)
        except subprocess.CalledProcessError:
            raise BuildException("Failed to build %s" % str(os.path.basename(sdist)))
        finally:
            xar_util.safe_rmtree(temp)

    def find(self, requirement):
        """
        Ensure all built wheels are added to the working set.
        Return the distribution.
        """
        finders.register_finders()
        for dist in pkg_resources.find_distributions(self._dest):
            if dist not in self._working_set:
                self._working_set.add(dist, entry=self._dest)
        return self._working_set.find(requirement)

    def __call__(self, requirement):
        """
        Attempts to download the requirement (and its dependencies) and add the
        wheel(s) to the downloads directory and the working set. Returns the
        distribution on success and None on failure.
        """
        # Remove non-wheels from the download directory
        self.clean()
        # Attempt to download the wheel/sdist
        try:
            self.download(requirement)
        except PipException as e:
            if self._log:
                self._log.exception(e)
            return None
        # Build wheels for the sdists (and remove the sdist)
        for entry in os.listdir(self._dest):
            if py_util.Wheel.is_wheel_archive(entry):
                continue
            try:
                sdist = os.path.join(self._dest, entry)
                self.build_wheel_from_sdist(sdist)
                xar_util.safe_remove(sdist)
            except BuildException as e:
                if self._log:
                    self._log.exception(e)
                return None
        # Return the wheel distribution
        return self.find(requirement)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import sys


PY2 = sys.version_info[0] == 2

try:
    # >= 3.4
    from importlib.util import cache_from_source, source_from_cache
except ImportError:
    if not PY2:
        # >= 3.0, < 3.4
        from imp import cache_from_source, source_from_cache
    else:
        # Not present in Python 2
        def cache_from_source(path, debug_override=None):
            assert path.endswith(".py")
            if debug_override is None:
                debug_override = __debug__
            if debug_override:
                suffix = "c"
            else:
                suffix = "o"
            return path + suffix

        def source_from_cache(path):
            assert path.endswith("c") or path.endswith("o")
            return path[:-1]


if PY2:

    def native(s, encoding="ascii"):
        return s


else:

    def native(s, encoding="ascii"):
        if isinstance(s, bytes):
            return s.decode(encoding)
        return s



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import tempfile
import unittest

from xar import py_util, xar_util


try:
    from unittest import mock
except ImportError:
    import mock


THISDIR = os.path.dirname(__file__)
TESTWHEEL = os.path.join(THISDIR, "test-1.0-py2.py3-none-win32.whl")


class PyUtilTest(unittest.TestCase):
    def test_environment_python_interpreter(self):
        interpreter = py_util.environment_python_interpreter()
        self.assertTrue(interpreter.startswith("/usr/bin/env"))

    def test_wheel_determine_kind_normal(self):
        distribution = mock.MagicMock(egg_info="xar-18.6.11-py3-none-any.whl")
        wheel = py_util.Wheel(distribution=distribution)
        lib = "/usr/lib/python/site-packages"
        paths = {
            "purelib": lib,
            "platlib": lib,
            "headers": "/usr/include/xar",
            "scripts": "/usr/bin",
            "data": "/usr",
        }
        k, p = wheel._determine_kind(lib, paths, paths, lib + "/tmp")
        self.assertTrue(k == "purelib" or k == "platlib")
        self.assertEqual(p, lib)
        k, p = wheel._determine_kind(lib, paths, paths, "/usr/none")
        self.assertEqual(k, "data")
        self.assertEqual(p, "/usr")
        k, p = wheel._determine_kind(lib, paths, paths, "/usr/bin/make_xar")
        self.assertEqual(k, "scripts")
        self.assertEqual(p, "/usr/bin")
        k, p = wheel._determine_kind(lib, paths, paths, "/usr/include/xar/x")
        self.assertEqual(k, "headers")
        self.assertEqual(p, "/usr/include/xar")

    def test_wheel_determine_kind_mac(self):
        distribution = mock.MagicMock(egg_info="xar-18.6.11-py3-none-any.whl")
        wheel = py_util.Wheel(distribution=distribution)
        lib = "/usr/lib/python/site-packages"
        paths = {
            "purelib": lib,
            "platlib": lib,
            "headers": "/bad/prefix/include/xar",
            "scripts": "/bad/prefix/bin",
            "data": "/bad/prefix",
        }
        k, p = wheel._determine_kind(lib, paths, paths, lib + "/tmp")
        self.assertTrue(k == "purelib" or k == "platlib")
        self.assertEqual(p, lib)
        k, p = wheel._determine_kind(lib, paths, paths, "/usr/none")
        self.assertEqual(k, "data")
        self.assertEqual(p, "/usr")
        k, p = wheel._determine_kind(lib, paths, paths, "/usr/bin/make_xar")
        self.assertEqual(k, "scripts")
        self.assertEqual(p, "/usr/bin")
        k, p = wheel._determine_kind(lib, paths, paths, "/usr/include/xar/x")
        self.assertEqual(k, "headers")
        self.assertEqual(p, "/usr/include/xar")

    def test_does_sha256_match(self):
        expected = "sha256=uU0nuZNNPgilLlLX2n2r-sSE7-N6U4DukIj3rOLvzek"
        unexpected = "sha256=XARXARXARXARXARXARXARXARXARXARXARXARXARXARX"
        with tempfile.NamedTemporaryFile("wb", delete=False) as f:
            f.write(b"hello world")
            file = f.name
        self.assertTrue(py_util.does_sha256_match(file, expected))
        self.assertFalse(py_util.does_sha256_match(file, unexpected))
        self.assertFalse(py_util.does_sha256_match(file, ""))

    def _temppaths(self, root):
        return {
            "purelib": root,
            "platlib": root,
            "headers": os.path.join(root, "include/xar"),
            "scripts": os.path.join(root, "bin"),
            "data": root,
        }

    def test_wheel_copy_installation(self):
        dist = mock.MagicMock()
        dist.location = "/path/to/lib/xar"
        dist.egg_info = "/path/to/lib/xar-18.7.12.dist-info"
        wheel = py_util.Wheel(distribution=dist)

        wheel.is_purelib = mock.MagicMock(return_value=True)
        wheel.records = mock.MagicMock(
            return_value=[
                ["xar/0", "sha256=uU0nuZNNPgilLlLX2n2r-sSE7-N6U4DukIj3rOLvzek", 11],
                ["xar/1", "", 11],
                ["xar/2", "sha256=uU0nuZNNPgilLlLX2n2r-sSE7-N6U4DukIj3rOLvzek", 11],
                ["xar/2", "sha256=uU0nuZNNPgilLlLX2n2r-sSE7-N6U4DukIj3rOLvzek", 11],
            ]
        )

        src = tempfile.mkdtemp()
        os.mkdir(os.path.join(src, "xar"))
        dst = tempfile.mkdtemp()

        for file, _, _ in wheel.records():
            with open(os.path.join(src, file), "wb") as f:
                f.write(b"hello world")

        wheel.copy_installation(self._temppaths(src), self._temppaths(dst))

        for file, hash, _ in wheel.records():
            dst_file = os.path.join(dst, file)
            self.assertTrue(os.path.exists(dst_file))
            matches = py_util.does_sha256_match(dst_file, hash)
            if hash:
                self.assertTrue(matches)
            else:
                self.assertFalse(matches)

        xar_util.safe_rmtree(src)
        xar_util.safe_rmtree(dst)

    def test_wheel_sys_install_paths(self):
        print(TESTWHEEL)
        wheel = py_util.Wheel(location=TESTWHEEL)
        sys_paths = wheel.sys_install_paths()
        for path in self._temppaths(""):
            self.assertTrue(path in sys_paths)

    def test_wheel_is_purelib(self):
        print(TESTWHEEL)
        wheel = py_util.Wheel(location=TESTWHEEL)
        self.assertFalse(wheel.is_purelib())

    def _check_install(self, paths):
        def assertExists(*path):
            self.assertTrue(os.path.exists(os.path.join(*path)))

        assertExists(paths["platlib"], "hello.pyd")
        assertExists(paths["platlib"], "hello", "hello.py")
        assertExists(paths["platlib"], "hello", "__init__.py")
        assertExists(paths["data"], "hello.dat")
        assertExists(paths["headers"], "hello.dat")
        assertExists(paths["scripts"], "hello.sh")
        assertExists(paths["platlib"], "test-1.0.dist-info", "RECORD")

    def test_wheel_install_archive(self):
        print(TESTWHEEL)
        self.assertTrue(py_util.Wheel.is_wheel_archive(TESTWHEEL))
        wheel = py_util.Wheel(location=TESTWHEEL)
        dst = tempfile.mkdtemp()
        paths = self._temppaths(dst)

        # Install without force
        wheel.install_archive(paths, force=False)
        self._check_install(paths)

        # Reinstall with force
        wheel.install_archive(paths, force=True)
        self._check_install(paths)

        xar_util.safe_rmtree(dst)

    def test_wheel_install(self):
        archived_wheel = py_util.Wheel(location=TESTWHEEL)
        src = tempfile.mkdtemp()
        dst = tempfile.mkdtemp()
        src_paths = self._temppaths(src)
        dst_paths = self._temppaths(dst)

        # Install the archive to src
        archived_wheel.install(None, src_paths)
        self._check_install(src_paths)

        # Copy the installation to dst
        src_location = os.path.join(src, "test-1.0.dist-info")
        self.assertFalse(py_util.Wheel.is_wheel_archive(src_location))
        installed_wheel = py_util.Wheel(location=src_location)
        installed_wheel.install(src_paths, dst_paths, force=False)
        self._check_install(dst_paths)

        # Reinstall copy to dst with force
        installed_wheel.install(src_paths, dst_paths, force=True)
        self._check_install(dst_paths)

        xar_util.safe_rmtree(src)
        xar_util.safe_rmtree(dst)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import contextlib
import os
import subprocess
import tempfile
import unittest

from xar import make_xar, xar_util


try:
    import zipapp
except ImportError:
    zipapp = None


class MakeXarTest(unittest.TestCase):
    @contextlib.contextmanager
    def make_test_directory(self):
        try:
            dir = tempfile.mkdtemp()
            with open(os.path.join(dir, "__main__.py"), "w") as f:
                f.write("print('python')")
            with open(os.path.join(dir, "other.py"), "w") as f:
                f.write("print('other')")
            with open(os.path.join(dir, "__main__.sh"), "w") as f:
                f.write("#!/bin/sh\necho shell")
            with tempfile.NamedTemporaryFile("w") as xarfile:
                yield xarfile.name, dir
        finally:
            xar_util.safe_rmtree(dir)

    def xarexec_exists(self):
        try:
            subprocess.check_call(["which", "xarexec_fuse"])
            return True
        except subprocess.CalledProcessError:
            return False

    def check_xar(self, xarfile):
        return subprocess.check_output([xarfile]).strip()

    def test_make_python_xar_from_directory(self):
        with self.make_test_directory() as (xar, dir):
            args = ["--output", xar, "--python", dir]
            make_xar.main(args)
            if self.xarexec_exists():
                self.assertEqual(self.check_xar(xar), b"python")
            args += ["--python-entry-point", "other"]
            make_xar.main(args)
            if self.xarexec_exists():
                self.assertEqual(self.check_xar(xar), b"other")

    def test_make_python_xar_from_archive(self):
        if zipapp is None:
            return
        with self.make_test_directory() as (xar, dir):
            with tempfile.NamedTemporaryFile("w") as zip:
                zipapp.create_archive(dir, zip.name)
                args = ["--output", xar, "--python", zip.name]
                make_xar.main(args)
                if self.xarexec_exists():
                    self.assertEqual(self.check_xar(xar), b"python")
                args += ["--python-entry-point", "other"]
                make_xar.main(args)
                if self.xarexec_exists():
                    self.assertEqual(self.check_xar(xar), b"other")

    def test_make_raw_xar(self):
        with self.make_test_directory() as (xar, dir):
            args = ["--output", xar, "--raw", dir]
            make_xar.main(args)
            if self.xarexec_exists():
                with self.assertRaises(OSError):
                    self.check_xar(xar)
            args += ["--raw-executable", "__main__.sh"]
            make_xar.main(args)
            if self.xarexec_exists():
                self.assertEqual(self.check_xar(xar), b"shell")



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import io
import os
import tempfile
import unittest

from xar import py_util, xar_builder, xar_util
from xar.tests import xar_test_helpers


class XarUtilTest(xar_test_helpers.XarTestCase):
    def test_xar_factory(self):
        "Test XarFactory and XarReader"

        # Make a boring xar file.
        srcdir = self.make_test_skeleton()

        tf = tempfile.NamedTemporaryFile(delete=False)
        xar = xar_util.XarFactory(srcdir, tf.name, xar_builder.BORING_SHEBANG)
        xar.squashfs_options.compression_algorithm = "gzip"
        xar.squashfs_options.block_size = 4096
        xar.go()

        outdir = os.path.join(tempfile.mkdtemp(), "squashfs-root")
        self._unxar(tf.name, outdir)
        self.assertDirectoryEqual(srcdir, outdir)

    def test_partition_files(self):
        "Test the file partitioning functionality used for split XARs."
        srcdir = xar_util.StagingDirectory(self.make_test_skeleton())
        dstdir = copy.deepcopy(srcdir)
        debuginfo_dir = xar_util.StagingDirectory()
        mp3_dir = xar_util.StagingDirectory()

        # Set up and execute the partitioning.
        print(
            "Partitioning %s (normal) to %s (debuginfo) and %s (mp3)"
            % (dstdir.path(), debuginfo_dir.path(), mp3_dir.path())
        )

        uuid = xar_util.make_uuid()
        extension_map = {
            ".debuginfo": xar_util.PartitionDestination(debuginfo_dir, uuid),
            ".mp3": xar_util.PartitionDestination(mp3_dir, uuid),
        }
        xar_util.partition_files(dstdir, extension_map)

        # Every debuginfo file in dstdir should be a symlink; every
        # .txt file should be a real file.  We should find the same
        # number of txt files as debuginfo symlinks.
        num_normal_files = 0
        num_symlinks = 0
        for dirname, _, filenames in os.walk(dstdir.path()):
            for filename in filenames:
                fn = os.path.join(dirname, filename)
                if fn.endswith((".debuginfo", ".mp3")):
                    self.assertTrue(os.path.islink(fn))
                    link = os.readlink(fn)
                    self.assertTrue(
                        link.find("/%s/" % uuid) != -1,
                        "%s symlink to %s contains /%s/" % (link, fn, uuid),
                    )
                    num_symlinks += 1
                else:
                    self.assertTrue(os.path.isfile(fn), "%s is a file" % fn)
                    num_normal_files += 1

        # Two symlinks per normal file
        self.assertEquals(num_normal_files, 7)
        self.assertEquals(2 * num_normal_files, num_symlinks)

        # Make sure only normal files are in the debuginfo dir.
        for dirname, _, filenames in os.walk(debuginfo_dir.path()):
            for filename in filenames:
                fn = os.path.join(dirname, filename)
                if fn.endswith(".debuginfo"):
                    self.assertTrue(os.path.isfile(fn))
                else:
                    self.fail("found non-debuginfo file in debug partition")

        # Same, but for mp3.
        for dirname, _, filenames in os.walk(mp3_dir.path()):
            for filename in filenames:
                fn = os.path.join(dirname, filename)
                if fn.endswith(".mp3"):
                    self.assertTrue(os.path.isfile(fn))
                else:
                    self.fail("found non-mp3 file in mp3 partition")

        self.assertDirectoryEqual(srcdir.path(), dstdir.path(), check_contents=False)

        srcdir.delete()
        dstdir.delete()
        debuginfo_dir.delete()
        mp3_dir.delete()

    def test_align_offset(self):
        self.assertEquals(0, xar_util._align_offset(0))
        self.assertEquals(4096, xar_util._align_offset(1))
        self.assertEquals(4096, xar_util._align_offset(4095))
        self.assertEquals(4096, xar_util._align_offset(4096))
        self.assertEquals(8192, xar_util._align_offset(4097))

    def test_long_header(self):
        """Test headers longer than 4096 bytes"""
        # Make a boring xar file.
        srcdir = self.make_test_skeleton()

        tf = tempfile.NamedTemporaryFile(delete=False)
        xar = xar_util.XarFactory(srcdir, tf.name, xar_builder.BORING_SHEBANG)
        xar.squashfs_options.compression_algorithm = "gzip"
        xar.squashfs_options.block_size = 4096
        xar.xar_header["IGNORED"] = "0" * 5000
        xar.go()

        outdir = os.path.join(tempfile.mkdtemp(), "squashfs-root")
        self._unxar(tf.name, outdir)
        self.assertDirectoryEqual(srcdir, outdir)

    def test_write_sort_file(self):
        """Tests write_sort_file()"""
        source_dir = self.make_test_skeleton()
        # Add in a file with spaces
        with open(os.path.join(source_dir, "space file.txt"), "w") as fh:
            fh.write("space file")

        sort_file = io.StringIO()
        priorities = [".txt", ".debuginfo", ".mp3"]
        xar_util.write_sort_file(source_dir, priorities, sort_file)
        sort_data = [
            line.split(" ") for line in sort_file.getvalue().strip().split("\n")
        ]
        for filename, priority in sort_data:
            self.assertFalse(" " in filename)
            if filename.endswith(".txt"):
                self.assertEqual(priority, "-4")
            if filename.endswith(".debuginfo"):
                self.assertEqual(priority, "-3")
            if filename.endswith(".mp3"):
                self.assertEqual(priority, "-2")

    def test_staging_deepcopy(self):
        original = xar_util.StagingDirectory(self.make_test_skeleton())
        clone = copy.deepcopy(original)
        self.assertNotEqual(original.absolute(), clone.absolute())
        self.assertDirectoryEqual(original.absolute(), clone.absolute())

    def test_temporary_file_deepcopy(self):
        original = xar_util.TemporaryFile()
        data = "the data"
        with original.open("w+t") as f:
            f.write(data)
        clone = copy.deepcopy(original)
        self.assertNotEqual(original.name(), clone.name())
        with clone.open("r+t") as f:
            self.assertEqual(data, f.read())

    def make_test_skeleton(self):
        "Make a simple tree of test files"
        srcdir = tempfile.mkdtemp()

        # First create a bunch of directories, and then files inside
        # each one -- one .txt and one .debuginfo file per directory.
        for d in "d1 d1/sub1 d1/sub2 d2 d3 d3/sub4".split():
            os.mkdir(os.path.join(srcdir, d))

        n = 0
        for dirpath, _, _ in os.walk(srcdir):
            with open(os.path.join(dirpath, "%s.txt" % n), "w") as fh:
                fh.write("orignal file %s\n" % n)
            n += 1

        for dirpath, _, _ in os.walk(srcdir):
            with open(os.path.join(dirpath, "%s.debuginfo" % n), "w") as fh:
                fh.write("debuginfo %s\n" % n)
            n += 1

        for dirpath, _, _ in os.walk(srcdir):
            with open(os.path.join(dirpath, "%s.mp3" % n), "w") as fh:
                fh.write("mp3 %s\n" % n)
            n += 1

        return srcdir


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from __future__ import absolute_import, division, print_function, unicode_literals

import os

from setuptools import setup


CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_long_description():
    with open(os.path.join(CURRENT_DIR, "README.md"), "r") as ld_file:
        return ld_file.read()


setup(
    name="xar",
    version="19.4.22",
    description="The XAR packaging toolchain.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Chip Turner",
    author_email="chip@fb.com",
    url="https://github.com/facebookincubator/xar",
    license="BSD",
    packages=[
        "xar",
        "xar.commands",
        "xar.tests",
        "xar.vendor",
        "xar.vendor.wheel",
        "xar.vendor.wheel.signatures",
    ],
    install_requires=[
        "pip>=10.0.1",
        # Version 34.1 fixes a bug in the dependency resolution. If this is
        # causing an problem for you, please open an issue, and we can evaluate
        # a workaround. (grep setuptools>=34.1 to see issue)
        # https://github.com/pypa/setuptools/commit/8c1f489f09434f42080397367b6491e75f64d838  # noqa: E501
        "setuptools>=34.1",
    ],
    test_requires=["mock", "pytest"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Framework :: Setuptools Plugin",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
    ],
    entry_points={
        "distutils.commands": ["bdist_xar = xar.commands.bdist_xar:bdist_xar"],
        "console_scripts": ["make_xar = xar.make_xar:main"],
    },
)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function

import os
import shutil
import subprocess
import sys
import tempfile
import unittest

import pkg_resources
from xar import pip_installer, xar_util


try:
    from unittest import mock
except ImportError:
    import mock


HELLO_SETUP_PY = """
from setuptools import setup
import wheel

setup(
    name="hello",
    version="0.0.0",
    packages=["hello"],
    entry_points={
        "console_scripts": ["hello = hello"],
    },
)
"""


class PyUtilTest(unittest.TestCase):
    def setUp(self):
        self.src = tempfile.mkdtemp()
        setup_py = os.path.join(self.src, "setup.py")
        xar_util.safe_mkdir(os.path.join(self.src, "hello"))
        with open(os.path.join(self.src, "README"), "w") as f:
            f.write("hello\n")
        with open(os.path.join(self.src, "hello/__init__.py"), "w") as f:
            f.write("print('hello')\n")
        with open(setup_py, "w") as f:
            f.write(HELLO_SETUP_PY)

        subprocess.check_call([sys.executable, setup_py, "bdist_wheel"], cwd=self.src)
        subprocess.check_call([sys.executable, setup_py, "sdist"], cwd=self.src)

        dist_dir = os.path.join(self.src, "dist")
        dists = os.listdir(dist_dir)
        self.assertEqual(len(dists), 2)
        if dists[0].lower().endswith(".whl"):
            self.wheel = os.path.join(dist_dir, dists[0])
            self.sdist = os.path.join(dist_dir, dists[1])
        else:
            self.wheel = os.path.join(dist_dir, dists[1])
            self.sdist = os.path.join(dist_dir, dists[0])
        self.req = pkg_resources.Requirement("hello")
        self.dst = tempfile.mkdtemp()

    def tearDown(self):
        xar_util.safe_rmtree(self.src)
        xar_util.safe_rmtree(self.dst)

    def mock_download_sdist(self, _req):
        shutil.copy(self.sdist, self._dest)

    def mock_download_wheel(self, _req):
        shutil.copy(self.wheel, self._dest)

    @mock.patch.object(pip_installer.PipInstaller, "download", mock_download_wheel)
    def test_pip_install_wheel(self):
        working_set = pkg_resources.WorkingSet(sys.path)
        installer = pip_installer.PipInstaller(self.dst, working_set)
        installer.sdist = self.sdist
        installer.wheel = self.wheel
        dist = installer(self.req)
        self.assertTrue(dist in self.req)

    @mock.patch.object(pip_installer.PipInstaller, "download", mock_download_sdist)
    def test_pip_install_sdist(self):
        working_set = pkg_resources.WorkingSet(sys.path)
        installer = pip_installer.PipInstaller(self.dst, working_set)
        installer.sdist = self.sdist
        installer.wheel = self.wheel
        dist = installer(self.req)
        self.assertTrue(dist in self.req)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals


BOOTSTRAP_XAR = "bootstrap_xar.sh"
RUN_XAR_MAIN = "__run_xar_main__.py"

BOOTSTRAP_XAR_TEMPLATE = """#!/bin/sh -eu

readlink_e() {{
    local path="$1"
    readlink -e "$path" 2>/dev/null && return

    # macosx / bsd readlink doesn't support -e
    # so use pwd -P with some recursive readlinking

    # strip trailing /
    path="${{path%/}}"

    # make path an absolute path
    if [[ "${{path:0:1}}" != "/" ]]
    then
        path="$(pwd -P)/$path"
    fi

    local slash_basename=""
    local counter=0
    while [[ -h "$path" ]]
    do
        if [[ counter -gt 200 ]]
        then
            echo "ERROR: Cyclical symbolic link detected: $path" 1>&2
            return
        fi
        counter=$(( counter + 1 ))

        target="$(readlink "$path")"
        if [[ "${{target:0:1}}" == "/" ]]
        then
            path="$target"
        else
            slash_basename="/$(basename "$path")"
            path="${{path%$slash_basename}}/$target"
        fi
    done

    # determine the target
    slash_basename="/$(basename "$path")"
    if [[ "$slash_basename" == "/.." || "$slash_basename" == "/." ]]
    then
        slash_basename=""
    fi
    local parent_dir="${{path%$slash_basename}}"

    # subshell to preserve the cwd (instead of pushd/popd)
    (cd "$parent_dir"; echo "$(pwd -P)$slash_basename")
}}

BOOTSTRAP_PATH="$0"
ORIGINAL_EXECUTABLE="$1"; shift
DIR=$(dirname "$BOOTSTRAP_PATH")

# Save any existing LD_LIBRARY_PATH
if [ -n "${{LD_LIBRARY_PATH+SET}}" ]; then
  export XAR_SAVED_LD_LIBRARY_PATH=$LD_LIBRARY_PATH
fi

# Don't inherit PYTHONPATH.  We set it to be the XAR mountpoint.
if [ -n "${{PYTHONPATH+SET}}" ]; then
  export XAR_SAVED_PYTHONPATH=$PYTHONPATH
fi

export XAR_INVOKED_NAME="$ORIGINAL_EXECUTABLE"
export LD_LIBRARY_PATH="$DIR"
export PYTHONPATH="$DIR"
export XAR_RUNTIME_FILES
XAR_RUNTIME_FILES="$(dirname "$(readlink_e "$BOOTSTRAP_PATH")")"
export XAR_PYTHON_COMMAND="{python}"

exec {python} "$DIR/{run_xar_main}" "$@"
"""


def run_xar_main(**kwargs):
    """
    Constructs the run_xar_main given the template arguments.
    If the {function} template argument is present, then the entry point
    {module}.{function}() is executed as main. Otherwise, {module} is run as the
    main module.
    """
    run_xar_main = """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# Put everything inside an __invoke_main() function.
# This way anything we define won't pollute globals(), since runpy
# will propagate our globals() as to the user's main module.
def __invoke_main():
    import fcntl
    import os
    import shlex
    import sys

    sys.argv[0] = os.getenv("XAR_INVOKED_NAME")

    # Hold a file descriptor open to a file inside our XAR to keep it
    # mounted while the xar is running.  We simply open the actual
    # directory rather than any file (which would also work).
    xar_mountpoint = os.getenv('XAR_RUNTIME_FILES')
    if xar_mountpoint:
        fd = os.open(xar_mountpoint, os.O_RDONLY)
        flags = fcntl.fcntl(fd, fcntl.F_GETFD)
        fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
"""
    if "function" in kwargs:
        run_xar_main += """
    import {module}
    {module}.{function}()
"""
    else:
        run_xar_main += """
    import runpy
    module = "{module}"
    runpy._run_module_as_main(module, False)
"""
    run_xar_main += """

__invoke_main()
"""

    return run_xar_main.format(**kwargs)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function

import os
import pkgutil
import zipimport

import pkg_resources
from xar import py_util


try:
    import importlib.machinery as importlib_machinery

    # access attribute to force import under delayed import mechanisms.
    importlib_machinery.__name__
except ImportError:
    importlib_machinery = None


def find_wheels_in_zip(importer, path_item, only=False):
    try:
        yield py_util.Wheel(location=path_item, importer=importer).distribution
    except Exception:
        pass


def find_wheels_on_path(importer, path_item, only=False):
    if only or not os.path.isdir(path_item) or not os.access(path_item, os.R_OK):
        return
    for entry in os.listdir(path_item):
        if py_util.Wheel.is_wheel_archive(entry):
            location = os.path.join(path_item, entry)
            for dist in pkg_resources.find_distributions(location):
                yield dist


def find_on_path(importer, path_item, only=False):
    for finder in (pkg_resources.find_on_path, find_wheels_on_path):
        for dist in finder(importer, path_item, only):
            yield dist


__REGISTERED = False


def register_finders():
    """
    Register pkg_resources finders that work with wheels (but not eggs). These
    replace the default pkg_resources finders. This function should be called
    before calling pkg_resources.find_distributions().
    """
    global __REGISTERED
    if __REGISTERED:
        return

    pkg_resources.register_finder(zipimport.zipimporter, find_wheels_in_zip)
    pkg_resources.register_finder(pkgutil.ImpImporter, find_on_path)
    if hasattr(importlib_machinery, "FileFinder"):
        pkg_resources.register_finder(importlib_machinery.FileFinder, find_on_path)

    __REGISTERED = True
